==> mo_gym-0.1.1-py3-none-any.whl: mo_gym/breakable_bottles/breakable_bottles.py <==
from gym import Env
from gym.spaces import Dict, Discrete, MultiBinary, Box
import numpy as np


class BreakableBottles(Env):
    metadata = {"render_modes": ["human", "rgb_array"]}

    # actions
    LEFT = 0
    RIGHT = 1
    PICKUP = 2

    def __init__(self, size=5, prob_drop=0.1, time_penalty=-1, bottle_reward=25, unbreakable_bottles=False, seed=None):
        # settings
        self.prob_drop = prob_drop
        self.time_penalty = time_penalty
        self.bottle_reward = bottle_reward
        self.unbreakable_bottles = False 

        # properties
        self.num_objectives = 3

        # initialize env state
        self.size = size
        self.location = 0
        self.bottles_carrying = 0
        self.bottles_delivered = 0
        self.bottles_dropped = [0]*(self.size - 2)

        # observation and action space
        self.observation_space = Dict(
            {
                "location": Discrete(self.size),
                "bottles_carrying": Discrete(3),
                "bottles_delivered": Discrete(2),
                "bottles_dropped": MultiBinary(self.size - 2),
            }
        )
        self.num_observations = 240

        self.action_space = Discrete(3) # LEFT, RIGHT, PICKUP
        self.num_actions = 3

        # reward space
        self.reward_space = Box(np.array([-np.inf, 0, -1]), np.array([0, self.bottle_reward*2, 0]))

    def step(self, action):
        observation_old = self._get_obs()
        old_potential = self.potential(observation_old)
        terminal = False
        reward = [self.time_penalty, 0, 0]

        if action == self.LEFT and self.location > 0:
            # execute bottle drop, if agent is carrying at least two and current location is 1, 2 or 3
            if self.location in range(1, self.size - 1) and self.bottles_carrying > 1 and np.random.random() < self.prob_drop:
                self.bottles_carrying -= 1
                self.bottles_dropped[self.location - 1] += 1

            # move to the left
            self.location -= 1

        elif action == self.RIGHT and self.location < self.size - 1:
            # execute bottle drop, if agent is carrying at least two and current location is 1, 2 or 3
            if self.location in range(1, self.size - 1) and self.bottles_carrying > 1 and np.random.random() < self.prob_drop:
                self.bottles_carrying -= 1
                self.bottles_dropped[self.location - 1] += 1

            # move to the right
            self.location += 1

            # if agent enters destination tile, deliver any carried bottles
            if self.location == self.size - 1 and self.bottles_carrying > 0:
                num_before = self.bottles_delivered
                num_after = min(self.bottles_delivered + self.bottles_carrying, 2)
                num_delivered = num_after - num_before
                self.bottles_delivered = num_after
                self.bottles_carrying = 0
                reward[1] += self.bottle_reward*num_delivered
                if self.bottles_delivered == 2:
                    terminal = True

        elif action == self.PICKUP:
            # agent is at source and has not reached carrying limit
            if self.location == 0 and self.bottles_carrying < 2:
                # add bottle to inventory
                self.bottles_carrying += 1
            # agent is on a tile where a dropped bottle lies and it has not reached the carrying limit
            elif self.location in range(1, self.size - 1) and self.bottles_dropped[self.location - 1] > 0 and self.bottles_carrying < 2 and self.unbreakable_bottles:
                # remove bottle from current tile
                self.bottles_dropped[self.location - 1] -= 1
                # add bottle to agent's inventory
                self.bottles_carrying += 1

        # next observation
        observation = self._get_obs()

        # calculate potential-based low impact measure
        #r2_t = phi(S_t) - phi(S_t-1)
        #sum_t(r2_t) = 0 -> no impact
        reward[2] = self.potential(observation) - old_potential

        info = {}
        return observation, reward, terminal, info
    
    def reset(self, seed=None, return_info=False, **kwargs):
        super().reset(seed=seed)
        self.np_random.seed(seed)
        self.r_star = 0
        self.location = self.size - 1
        self.bottles_carrying = 0
        self.bottles_delivered = 0
        self.bottles_dropped = [0]*(self.size - 2)
        state = self._get_obs()
        return (state, {}) if return_info else state
    
    def get_obs_idx(self, obs):
        multi_index = np.array([[obs["location"]],
                                [obs["bottles_carrying"]], 
                                [obs["bottles_delivered"]],
                                *[[bd > 0] for bd in obs["bottles_dropped"]]])
        return np.ravel_multi_index(multi_index, tuple([self.size, 3, 2, *([2]*(self.size - 2))]))
    
    def _get_obs(self):
        return {"location": self.location,
                "bottles_carrying": self.bottles_carrying,
                "bottles_delivered": self.bottles_delivered,
                "bottles_dropped": self.bottles_dropped.copy()}
    
    def render(self, mode="human"):
        if mode == 'rgb_array':
            return np.array([self.state[:3], *self.state[3]]) # return RGB frame suitable for video
        elif mode == 'human':
            print("-----")
            print(f"Location: {self.location}\nCarrying {self.bottles_carrying} bottles.\nDelivered {self.bottles_delivered} so far.\nBottles have been dropped at tiles {'1' if self.bottles_dropped[0] > 0 else ''} {'2' if self.bottles_dropped[1] > 0 else ''} {'3' if self.bottles_dropped[2] > 0 else ''}")
            print("-----")
        else:
            super(BreakableBottles, self).render(mode=mode) # just raise an exception

    def close(self):
        pass

    def seed(self, seed=None):
        self.seed = seed if not seed is None else np.random.randint(2**32) 

    def potential(self, obs):
        if sum(obs["bottles_dropped"]) > 0:
            return -1
        return 0

if __name__ == '__main__':
    from gym.spaces.utils import flatdim
    env = BreakableBottles(size=5, prob_drop=0.1)
    assert flatdim(env.action_space) == 3
    assert flatdim(env.observation_space) == 13

    done = False
    obs = env.reset()
    while True:
        env.render()
        obs, r, done, info = env.step(env.action_space.sample())
        if done:
            break

==> mo_gym-0.2.0-py3-none-any.whl: mo_gym/breakable_bottles/breakable_bottles.py <==
from os import terminal_size
from typing import Optional
from gym import Env
from gym.spaces import Dict, Discrete, MultiBinary, Box
import numpy as np


class BreakableBottles(Env):
    metadata = {"render_modes": ["human"]}

    # actions
    LEFT = 0
    RIGHT = 1
    PICKUP = 2

    def __init__(self, render_mode: Optional[str] = None, size=5, prob_drop=0.1, time_penalty=-1, bottle_reward=25, unbreakable_bottles=False):
        self.render_mode = render_mode
        
        # settings
        self.prob_drop = prob_drop
        self.time_penalty = time_penalty
        self.bottle_reward = bottle_reward
        self.unbreakable_bottles = False 

        # properties
        self.num_objectives = 3

        # initialize env state
        self.size = size
        self.location = 0
        self.bottles_carrying = 0
        self.bottles_delivered = 0
        self.bottles_dropped = [0]*(self.size - 2)

        # observation and action space
        self.observation_space = Dict(
            {
                "location": Discrete(self.size),
                "bottles_carrying": Discrete(3),
                "bottles_delivered": Discrete(2),
                "bottles_dropped": MultiBinary(self.size - 2),
            }
        )
        self.num_observations = 240

        self.action_space = Discrete(3) # LEFT, RIGHT, PICKUP
        self.num_actions = 3

        # reward space
        self.reward_space = Box(np.array([-np.inf, 0, -1]), np.array([0, self.bottle_reward*2, 0]))

    def step(self, action):
        observation_old = self._get_obs()
        old_potential = self.potential(observation_old)
        terminal = False
        reward = [self.time_penalty, 0, 0]

        if action == self.LEFT and self.location > 0:
            # execute bottle drop, if agent is carrying at least two and current location is 1, 2 or 3
            if self.location in range(1, self.size - 1) and self.bottles_carrying > 1 and np.random.random() < self.prob_drop:
                self.bottles_carrying -= 1
                self.bottles_dropped[self.location - 1] += 1

            # move to the left
            self.location -= 1

        elif action == self.RIGHT and self.location < self.size - 1:
            # execute bottle drop, if agent is carrying at least two and current location is 1, 2 or 3
            if self.location in range(1, self.size - 1) and self.bottles_carrying > 1 and np.random.random() < self.prob_drop:
                self.bottles_carrying -= 1
                self.bottles_dropped[self.location - 1] += 1

            # move to the right
            self.location += 1

            # if agent enters destination tile, deliver any carried bottles
            if self.location == self.size - 1 and self.bottles_carrying > 0:
                num_before = self.bottles_delivered
                num_after = min(self.bottles_delivered + self.bottles_carrying, 2)
                num_delivered = num_after - num_before
                self.bottles_delivered = num_after
                self.bottles_carrying = 0
                reward[1] += self.bottle_reward*num_delivered
                if self.bottles_delivered == 2:
                    terminal = True

        elif action == self.PICKUP:
            # agent is at source and has not reached carrying limit
            if self.location == 0 and self.bottles_carrying < 2:
                # add bottle to inventory
                self.bottles_carrying += 1
            # agent is on a tile where a dropped bottle lies and it has not reached the carrying limit
            elif self.location in range(1, self.size - 1) and self.bottles_dropped[self.location - 1] > 0 and self.bottles_carrying < 2 and self.unbreakable_bottles:
                # remove bottle from current tile
                self.bottles_dropped[self.location - 1] -= 1
                # add bottle to agent's inventory
                self.bottles_carrying += 1

        # next observation
        observation = self._get_obs()

        # calculate potential-based low impact measure
        #r2_t = phi(S_t) - phi(S_t-1)
        #sum_t(r2_t) = 0 -> no impact
        reward[2] = self.potential(observation) - old_potential

        info = {}
        if self.render_mode == "human":
            self.render()
        return observation, reward, terminal, False, info
    
    def reset(self, *, seed: Optional[int] = None, options: Optional[dict] = None):
        super().reset(seed=seed)
        self.r_star = 0
        self.location = self.size - 1
        self.bottles_carrying = 0
        self.bottles_delivered = 0
        self.bottles_dropped = [0]*(self.size - 2)
        state = self._get_obs()
        if self.render_mode == "human":
            self.render()
        return state, {}
    
    def get_obs_idx(self, obs):
        multi_index = np.array([[obs["location"]],
                                [obs["bottles_carrying"]], 
                                [obs["bottles_delivered"]],
                                *[[bd > 0] for bd in obs["bottles_dropped"]]])
        return np.ravel_multi_index(multi_index, tuple([self.size, 3, 2, *([2]*(self.size - 2))]))
    
    def _get_obs(self):
        return {"location": self.location,
                "bottles_carrying": self.bottles_carrying,
                "bottles_delivered": self.bottles_delivered,
                "bottles_dropped": self.bottles_dropped.copy()}
    
    def render(self):
        if self.render_mode == 'human':
            print("-----")
            print(f"Location: {self.location}\nCarrying {self.bottles_carrying} bottles.\nDelivered {self.bottles_delivered} so far.\nBottles have been dropped at tiles {'1' if self.bottles_dropped[0] > 0 else ''} {'2' if self.bottles_dropped[1] > 0 else ''} {'3' if self.bottles_dropped[2] > 0 else ''}")
            print("-----")

    def close(self):
        pass

    def potential(self, obs):
        if sum(obs["bottles_dropped"]) > 0:
            return -1
        return 0

if __name__ == '__main__':
    from gym.spaces.utils import flatdim
    env = BreakableBottles(size=5, prob_drop=0.1)
    assert flatdim(env.action_space) == 3
    assert flatdim(env.observation_space) == 13

    done = False
    obs = env.reset()
    while True:
        env.render()
        obs, r, done, info = env.step(env.action_space.sample())
        if done:
            break
